from .shape3d import Shape3D
from .infinite_shape import InfiniteShape

from .composition import Composition

from .cuboid import Cuboid
from .ellipsoid import Ellipsoid
from .tunnel import Tunnel
from .prism import Prism
from .obj2 import Obj2
from .bar_framework import BarFramework
from .plate_framework import PlateFramework
from .pillar_framework import PillarFramework

from .full_space import FullSpace
from .surface_topo import SurfaceTopo
Terrain = SurfaceTopo
