def some_func(x):
    return x


class Clazz:
    def __init__(self):
        pass


class DummyClass:
    def __init__(self):
        print('Dummy here')


class ValueHolder:
    def __init__(self, x=1):
        self.x = x
