from .simpeg_patch_context import simpeg_patched

from .tied import Tied
from .distributed import Distributed
from .progressed import Progressed
from .demag import Demaged

# try:
#     from .torched import Torched
# except ImportError:
#     pass  # most likely torch is not installed
